import numpy as np
import matplotlib.pyplot as plt

class GaussianCrossPSF(object):
    """ 
    Class to generate star-like pixel art using Gaussian function and masking.
    """
    def __init__(self, amplitude=1):
        """
        Initialize the GaussianCrossPSF class

        Args:
            amplitude (float): Amplitude of the Gaussian function
        
        """
        self.amplitude = amplitude

    def generate_cross_psf(self, x_center, y_center, stddev, background_factor, size=(300, 200)):
        """
        Generate Stars using Gaussian

        Function to generate a star-like pixel art using gaussian function and masking.

        Args:
            x_center (float): x coordinate of the center of the star
            y_center (float): y coordinate of the center of the star
            stddev (float): Standard deviation of the Gaussian function to control the size
            background_factor (float): A number from 0 to 1 to tune the masking
            size (tuple): Size of the output image

        Returns:
            np.ndarray: A 2D numpy array representing the star
        
        """
        x = np.linspace(0, size[0]-1, size[0])
        y = np.linspace(0, size[1]-1, size[1])
        x, y = np.meshgrid(x, y)
        
        # Calculate the Gaussian
        psf = self.amplitude * np.exp(-((x - x_center)**2 + (y - y_center)**2) / (2 * stddev**2))
        
        # Calculate the Gaussian along the x and y axes
        psf_x = self.amplitude * np.exp(-((x - x_center)**2) / (2 * stddev**2))
        psf_y = self.amplitude * np.exp(-((y - y_center)**2) / (2 * stddev**2))
        
        # Create a mask that retains higher values along the x and y axes
        mask_x = np.abs(y - y_center) <= 1e-5
        mask_y = np.abs(x - x_center) <= 1e-5
        
        # Apply the mask to the PSF values
        psf_cross = background_factor * psf
        psf_cross[mask_x] = psf_x[mask_x]
        psf_cross[mask_y] = psf_y[mask_y]
        
        return psf_cross

    def plot_multiple_cross_psfs(self, centers_stddevs, size=(300, 200)):
        """
        Plot Multiple Stars
        
        Function to plot multiple star-like pixel art using gaussian function and masking.

        Args:
            centers_stddevs (list): A list of tuples containing the x_center, y_center, stddev, and background_factor of each star
            size (tuple): Size of the output image

        """

        combined_psf = np.zeros((size[1], size[0]))
        
        for (x_center, y_center, stddev, background_factor) in centers_stddevs:
            psf_cross = self.generate_cross_psf(x_center, y_center, stddev, background_factor, size=size)
            combined_psf += psf_cross
        
        plt.figure(figsize=(6, 6))
        plt.imshow(combined_psf, extent=(0, size[0]-1, 0, size[1]-1), origin='lower')
        plt.colorbar()
        plt.xlabel('X')
        plt.ylabel('Y')
        plt.show()

# Example:
def main():
    psf = GaussianCrossPSF(amplitude=1)
    centers_stddevs = [(5, 5, 2.5, 0.5), (10, 20, 3, 0.4), (25, 10, 5, 0.4)]
    psf.plot_multiple_cross_psfs(centers_stddevs, size=(100, 67))

if __name__ == '__main__':
    main()



from astropy.nddata import Cutout2D
import numpy as np
from astropy.io import fits
from astropy.wcs import WCS
from spectral_cube import SpectralCube


def get_cutout_fn(filename, position, l, w, format='fits'):
    if format == 'fits':
        try: 
            hdu = fits.open(filename)
            data = np.squeeze(hdu['SCI'].data)
            head = hdu['SCI'].header
        except: 
            hdu = fits.open(filename)[0]
            data = np.squeeze(hdu.data)
            head = hdu.header
    elif format == 'casa':
        hdu = SpectralCube.read(filename, format='casa').hdu
        data = np.squeeze(hdu.data)
        head = hdu.header
    ww = WCS(head).celestial
    size = (l, w)
    cutout = Cutout2D(data, position=position, size=size, wcs=ww)
    return cutout

def get_cutout_hdu(hdu, position, l, w):
    data = np.squeeze(hdu.data)
    head = hdu.header
    ww = WCS(head).celestial
    size = (l, w)
    cutout = Cutout2D(data, position=position, size=size, wcs=ww)
    return cutout




from astropy import units as u
from astropy.coordinates import SkyCoord
from astroquery.vizier import Vizier
import numpy as np
import matplotlib.pyplot as plt

def get_catalog(catalog_name, coord, radius=1.0*u.arcmin, max_star=30):
    """ Get Catalog
    
    Function to get catalog at a given coordinate and radius using Vizier.

    Args:
        catalog_name (str): Name of the catalog.
        coord (SkyCoord): Coordinates of the center of the search.
        radius (Quantity): Radius of the search.

    Returns:
        Table: Table of the catalog.

    """
    try:
        guide = Vizier(catalog=catalog_name).query_region(coord, radius=radius)[0]
    except:
        print("No catalog found. Setting 2MASS catalog.")
        guide = get_2mass_catalog(coord, radius)

    if len(guide) > max_star:
        for c in guide.colnames:
            if 'mag' in c:
                guide = guide[~np.isnan(np.array(guide[c]))]
        if len(guide) <= max_star:
            return guide
        else:
            guide = guide[0:max_star]

    return guide

def get_2mass_catalog(coord, radius=1.0*u.arcmin):
    """ Query 2MASS Catalog
    
    Function to get 2MASS catalog at a given coordinate and radius using Vizier.

    Args:
        coord (SkyCoord): Coordinates of the center of the search.
        radius (Quantity): Radius of the search.

    Returns:
        Table: Table of the 2MASS catalog.

    """
    guide = Vizier(catalog="II/246").query_region(coord, radius=radius)[0]
    return guide

def get_sdss_catalog(coord, radius=1.0*u.arcmin):
    """ Query SDSS Catalog
    
    Function to get SDSS catalog at a given coordinate and radius using Vizier.

    Args:
        coord (SkyCoord): Coordinates of the center of the search.
        radius (Quantity): Radius of the search.

    Returns:
        Table: Table of the SDSS catalog.

    """
    guide = Vizier(catalog="V/147").query_region(coord, radius=radius)[0]
    return guide

def get_catalogs_list():
    return '2MASS', 'SDSS'

def get_random_coordinates():
    """ Get Random Coordinates

    Function to generate a random SkyCoord object.

    Returns:
        SkyCoord: Random SkyCoord object.
    """
    ra = np.random.uniform(0, 360)
    dec = np.random.uniform(-90, 90)
    return SkyCoord(ra=ra, dec=dec, unit=(u.deg, u.deg), frame='icrs')

def get_random_coordinates_gal():
    """ Get Random Coordinates in Galactic Coordinates

        Function to generate a random SkyCoord object in Galactic coordinates within -5 < b < 5.

        Returns:
            SkyCoord: Random SkyCoord object in Galactic coordinates within -5 < b < 5.
    """
    l = np.random.uniform(0, 360)
    b = np.random.uniform(-5, 5)
    return SkyCoord(l=l, b=b, unit=(u.deg, u.deg), frame='galactic')



import pytest
import numpy as np
import astropixel
from astropixel import catalog_querry
from astropixel import plot_stars
from astropixel import make_star
from astropixel.make_star import GaussianCrossPSF
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.table import Table

def test_generate_cross_psf():
    """ 
    Test the generate_cross_psf method of the GaussianCrossPSF class
    """
    psf = make_star.GaussianCrossPSF(amplitude=1)
    x_center = 150
    y_center = 100
    stddev = 10
    background_factor = 0.1
    size = (300, 200)
    psf_cross = psf.generate_cross_psf(x_center, y_center, stddev, background_factor, size=size)
    assert psf_cross.shape == (200, 300)
    assert np.max(psf_cross) == pytest.approx(1, abs=1e-3)
    assert np.min(psf_cross) == pytest.approx(0, abs=1e-3)
    assert np.argmax(psf_cross) == 30150

def test_get_catalog():
    """
    Test the get_catalog method of the catalog_querry module
    """
    coord = SkyCoord(ra=10.68458, dec=41.26917, unit=(u.deg, u.deg))
    radius = 1.0*u.arcmin
    catalog_name = '2MASS'
    guide = catalog_querry.get_catalog(catalog_name, coord, radius=radius)
    assert isinstance(guide, Table)
    assert len(guide) > 0

def test_get_random_coordinates():
    """ 
    Test the get_random_coordinates method of the catalog_querry module
    """
    coord = catalog_querry.get_random_coordinates()
    assert isinstance(coord, SkyCoord)
    assert coord.ra.value >= 0
    assert coord.ra.value <= 360
    assert coord.dec.value >= -90
    assert coord.dec.value <= 90

def test_get_random_coordinates_gal():
    """
    Test the get_random_coordinates_gal method of the catalog_querry module
    """
    coord = catalog_querry.get_random_coordinates_gal()
    assert isinstance(coord, SkyCoord)
    assert coord.l.value >= 0
    assert coord.l.value <= 360
    assert coord.b.value >= -5
    assert coord.b.value <= 5

def test_plot_field():
    """ 
    Test the StarPlotter class of the plot_stars module
    """
    coord = SkyCoord(ra=10.68458, dec=41.26917, unit=(u.deg, u.deg))
    field = plot_stars.StarPlotter(coord, radius=1.0*u.arcmin, size=(1000, 1000))
    assert field.size == (1000, 1000)
    assert field.radius == 1.0*u.arcmin
    assert field.coord == coord

def test_get_wcs():
    """ 
    Test the get_wcs method of the StarPlotter class
    """
    coord = SkyCoord(ra=10.68458, dec=41.26917, unit=(u.deg, u.deg))
    field = plot_stars.StarPlotter(coord, radius=1.0*u.arcmin, size=(1000, 1000))
    wcs = field.get_wcs()
    assert wcs.wcs.ctype[0] == 'RA---TAN'
    assert wcs.wcs.ctype[1] == 'DEC--TAN'
    assert wcs.wcs.crval[0] == coord.ra.value
    assert wcs.wcs.crval[1] == coord.dec.value

    pixcrd = np.array([[0, 0], [24, 38], [45, 98]], dtype=np.float64)
    world = wcs.all_pix2world(pixcrd, 0)
    ### This test is not working :(
    #assert world[0][0] == pytest.approx(coord.ra.value, abs=1e-3)
    #assert world[0][1] == pytest.approx(coord.dec.value, abs=1e-3)

    pixcrd2 = wcs.wcs_world2pix(world, 0)
    assert pixcrd2[0][0] == pytest.approx(0, abs=1e-3)
    assert pixcrd2[0][1] == pytest.approx(0, abs=1e-3)

    x = 0
    y = 0
    origin = 0
    assert (wcs.wcs_pix2world(x, y, origin) ==
            wcs.wcs_pix2world(x + 1, y + 1, origin + 1))

if __name__ == "__main__":
    test_generate_cross_psf()
    test_get_catalog()
    test_get_random_coordinates()
    test_get_random_coordinates_gal()
    test_plot_field()
    test_get_wcs()


from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.wcs import WCS
from astropy.visualization import simple_norm

import numpy as np

import matplotlib.pyplot as plt

from . import catalog_querry
from . import make_star


class StarPlotter(object):
    """ 
    Object for plotting stars
    """

    def __init__(self, coord, size=(1000, 1000), radius=1*u.arcmin, catalog_name='2MASS'):
        """ 
        Initialize the class

        Args:
            coord (SkyCoord): Coordinates of the center of the field.
            size (tuple): Size of the field in pixels.
            radius (Quantity): Radius of the field. Limited to < 3 arcmin due to catalog query limitations. 
                               If there are no stars in the field, try increasing the radius or changing the 
                               central coordinate.

        """

        self.coord = coord.icrs
        self.size = size
        self.size_scale = np.min(size)/10
        self.radius = radius
        self.scale = radius*2
        self.figsize = (5,4)
        self.R = None
        self.G = None
        self.B = None
        self.RA = None
        self.DEC = None
        self.catalog_name = catalog_name
        self.set_catalog(catalog_name)
        self.wcs = self.get_wcs()
        self.crosshair = False
    
    def set_catalog(self, catalog_name):
        """ 
        Function to get catalog

        Returns:
            Table: Table of stars.
        """
        self.catalog_name = catalog_name

        if catalog_name == '2MASS':
            self.cat = catalog_querry.get_2mass_catalog(self.coord, self.radius)
            self.R = 'Kmag'
            self.G = 'Hmag'
            self.B = 'Jmag'
            self.RA = 'RAJ2000'
            self.DEC = 'DEJ2000'
        elif catalog_name == 'SDSS':
            self.cat = catalog_querry.get_sdss_catalog(self.coord, self.radius)
            self.R = 'imag'
            self.G = 'gmag'
            self.B = 'umag'
            self.RA = 'RA_ICRS'
            self.DEC = 'DE_ICRS'
        else:
            self.cat = catalog_querry.get_catalog(catalog_name, self.coord, self.radius)
    
    def get_wcs(self):
        """ 
        Function to get WCS

        Returns:
            WCS: WCS object.
        """
        wcs = WCS(naxis=2)
        wcs.wcs.crpix = [self.size[0]/2, self.size[1]/2]  # Set the reference pixel to the center of the image
        wcs.wcs.crval = [self.coord.ra.deg, self.coord.dec.deg]  # Set the reference value to the given coordinates
        wcs.wcs.cdelt = np.array([-self.scale.to(u.deg).value/self.size[0], self.scale.to(u.deg).value/self.size[1]])  # Set the pixel scale 
        wcs.wcs.ctype = ["RA---TAN", "DEC--TAN"]  # Set the coordinate type to RA/DEC

        return wcs

    def plot_crosshair(ax):
        """
        Function to plot crosshair

        Args:
            ax (Axes): Axes object.

        Returns:
            ax (Axes): Axes object.
        """

        ax.scatter(self.coord.ra.deg, self.coord.dec.deg, s=100, edgecolor='r', facecolor='none')
        ax.scatter(self.coord.ra.deg, self.coord.dec.deg, s=100, marker='+', color='r')
        return ax

    def plot_scatter_field(self, labels=False, ax=None):
        """
        Function to plot field with scatter

        Args:
            labels (bool): Show labels.
            ax (Axes): Axes object.

        Returns:
            ax (Axes): Axes object.
        """

        if ax is None:
            fig = plt.figure(figsize=self.figsize)
            ax = fig.add_subplot(111, projection=self.wcs)

        ax.scatter(self.cat[self.RA], self.cat[self.DEC], s=50, transform=ax.get_transform('world'), marker='*', color='orange')
        if self.crosshair:
            ax = self.plot_crosshair(ax)

        if labels:
            ax.set_xlabel('Right Ascension')
            ax.set_ylabel('Declination')
        else:
            ax.set_axis_off()

        plt.tight_layout()
        return ax

    def get_cross_psf_field_image(self, band='Kmag'):
        """
        Function to get cross PSF field image

        Args:
            band (str): Band to use. Must match the column name in the catalog self.cat

        Returns:
            np.ndarray: Image of the field with cross PS
        """
        star = make_star.GaussianCrossPSF(amplitude=1)
        psf = np.zeros((self.size[1], self.size[0]))

        for c in self.cat:
            coordi = SkyCoord(c[self.RA], c[self.DEC], unit=(u.deg, u.deg), frame='icrs')
            pix_cord = coordi.to_pixel(self.wcs)
            std = scale_the_magnitude(c[band], scale=5)*self.size_scale
            psf += star.generate_cross_psf(np.round(pix_cord[0]), np.round(pix_cord[1]), std, 0.5, size=self.size)
        
        return psf

    def plot_cross_psf_field(self, band='Kmag', labels=False, ax=None):
        """
        Function to plot cross PSF field with imshow

        Args:
            labels (bool): Show labels.
            ax (Axes): Axes object.

        Returns:
            ax (Axes): Axes object.
        """
        if ax is None:
            fig = plt.figure(figsize=self.figsize)
            ax = fig.add_subplot(111, projection=self.wcs)

        psf = self.get_cross_psf_field_image(band=self.R)

        ax.imshow(psf, origin='lower', cmap='gray', aspect='equal')
        if labels:
            ax.set_xlabel('Right Ascension')
            ax.set_ylabel('Declination')
        else:
            ax.set_axis_off()
        
        if self.crosshair:
            ax = self.plot_crosshair(ax)
        
        return ax

    def plot_cross_psf_field_rgb(self, labels=False, ax=None):
        """
        Function to plot cross PSF field with imshow

        Args:
            labels (bool): Show labels.
            ax (Axes): Axes object.

        Returns:
            ax (Axes): Axes object.
        """
        if ax is None:
            fig = plt.figure(figsize=self.figsize)
            ax = fig.add_subplot(111, projection=self.wcs)

        #psf = self.get_cross_psf_field_image(band='Kmag')
        psf_R = self.get_cross_psf_field_image(band=self.R)
        psf_G = self.get_cross_psf_field_image(band=self.G)
        psf_B = self.get_cross_psf_field_image(band=self.B)
        
        rgb = make_rgb_scaled_image(psf_R, psf_G, psf_B)

        ax.imshow(rgb, origin='lower', aspect='equal')
        if labels:
            ax.set_xlabel('Right Ascension')
            ax.set_ylabel('Declination')
        else:
            ax.set_axis_off()

        if self.crosshair:
            ax = self.plot_crosshair(ax)
        
        return ax

def make_rgb_scaled_image(psf_R, psf_G, psf_B):
    """
    Function to make RGB image

    Args:
        psf_R (np.ndarray): Image of the field for the R color.
        psf_G (np.ndarray): Image of the field for the G color.
        psf_B (np.ndarray): Image of the field for the B color.

    Returns:
        np.ndarray: RGB image.
    """
    rgb = np.array(
        [
            psf_R,
            psf_G,
            psf_B
        ]
    ).swapaxes(0,2).swapaxes(0,1)

    rgb_scaled = np.array([
            simple_norm(rgb[:,:,0], stretch='asinh', vmin=0, vmax=1)(rgb[:,:,0]),
            simple_norm(rgb[:,:,1], stretch='asinh', vmin=0, vmax=1)(rgb[:,:,1]),
            simple_norm(rgb[:,:,2], stretch='asinh', vmin=0, vmax=1)(rgb[:,:,2]),
        ]
    ).swapaxes(0,2)

    return rgb_scaled.swapaxes(0,1)

def plot_random_scatter_field(size=(100, 100), radius=1*u.arcmin, ax=None):
    """
    Function to plot random field.

    Args:
        size (tuple): Size of the field in pixels.
        radius (Quantity): Radius of the field.
        ax (Axes): Axes object.

    Returns:
        ax (Axes): Axes object.
    """
    coord = catalog_querry.get_random_coordinates_gal()
    field = StarPlotter(coord, radius=radius, size=size)
    field.plot_scatter_field(ax=ax)

def plot_random_cross_psf_field(size=(100, 100), radius=1*u.arcmin, ax=None):
    """
    Function to plot random cross PSF field.

    Args:
        size (tuple): Size of the field in pixels.
        radius (Quantity): Radius of the field.
        ax (Axes): Axes object.

    Returns:
        ax (Axes): Axes object.
    """
    coord = catalog_querry.get_random_coordinates_gal()
    field = StarPlotter(coord, radius=radius, size=size)
    field.plot_cross_psf_field(ax=ax)

def plot_random_cross_psf_field_rgb(size=(100, 100), radius=1*u.arcmin, ax=None):
    """
    Function to plot random cross PSF field RGB.

    Args:
        size (tuple): Size of the field in pixels.
        radius (Quantity): Radius of the field.
        ax (Axes): Axes object.

    Returns:
        ax (Axes): Axes object.
    """
    coord = catalog_querry.get_random_coordinates_gal()
    field = StarPlotter(coord, radius=radius, size=size)
    field.plot_cross_psf_field_rgb(ax=ax)

def magnitude_to_luminosity(magnitude):
    """ 
    Function to convert magnitude to luminosity

    Args:
        magnitude (float): Magnitude of the star.

    Returns:
        float: Luminosity of the star

    """
    # Convert magnitude to flux
    flux = 10**(-0.4 * magnitude)
    
    # Convert flux to luminosity
    luminosity = 4 * np.pi * (10)**2 * flux
    
    return luminosity

def scale_the_magnitude(magnitude, scale=5):
    """ 
    Function to scale the magnitude

    Args:
        magnitude (float): Magnitude of the star.
        scale (float): Scale factor.

    Returns:
        float: Scaled magnitude.
    """
    lum = magnitude_to_luminosity(magnitude)

    return lum**(1/scale)

def example_plot_scatter_field(ax=None):
    """
    Example to plot scatter field

    Args:
        ax (Axes): Axes object.

    Returns:
        ax (Axes): Axes object.
    """
    coord = SkyCoord.from_name('Barnard\'s Star')
    field = StarPlotter(coord, size=(1000, 1000), radius=1*u.arcmin)

    if ax is None:
        fig = plt.figure(figsize=(5, 4))
        ax = plt.subplot(111, projection=field.wcs)

    ax = field.plot_scatter_field(ax=ax)
    plt.show()

def example_plot_cross_psf_field(ax=None):
    """ 
    Example to plot cross PSF field

    Args:
        ax (Axes): Axes object.

    Returns:
        ax (Axes): Axes object.
    """

    coord = SkyCoord.from_name('Barnard\'s Star')
    field = StarPlotter(coord, size=(1000, 1000), radius=1*u.arcmin)

    if ax is None:
        fig = plt.figure(figsize=(5, 4))
        ax = plt.subplot(111, projection=field.wcs)
    
    ax = field.plot_cross_psf_field(ax=ax)
    plt.show()

def example_plot_cross_psf_field_rgb():
    """
    Example to plot cross PSF field RGB
    """
    coord = SkyCoord.from_name('Barnard\'s Star')
    field = StarPlotter(coord, size=(50, 100), radius=1*u.arcmin)

    fig = plt.figure(figsize=(5, 4))
    ax = plt.subplot(111, projection=field.wcs)
    
    ax = field.plot_cross_psf_field_rgb(ax=ax)
    plt.show()

def main():
    #example_plot_cross_psf_field()
    example_plot_cross_psf_field_rgb()

if __name__ == '__main__':
    main()